import os
import itertools

from pymatgen.io.cif import CifParser
from pymatgen.io.vasp import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Structure
import pymatgen as pmg
import numpy as np

from .utils import is_file_type, is_not_file_type
from ..potential import Potential
from ..training import Training
from ..predict import Predict
from ..predict.utils import print_elastic_information
from ..io.minimal import MinimalMDReader
from ..visualize import visualize_single_calculation, visualize_radial_pair_distribution, visualize_pair_energies


def add_subcommand_test(subparsers):
    parser = subparsers.add_parser('test', help='train potential model')
    sub_subparsers = parser.add_subparsers()
    add_subcommand_test_properties(sub_subparsers)
    add_subcommand_test_relax(sub_subparsers)
    add_subcommand_test_training(sub_subparsers)
    add_subcommand_test_radial(sub_subparsers)
    add_subcommand_test_pair(sub_subparsers)


def add_subcommand_test_properties(subparsers):
    parser = subparsers.add_parser('properties', help='calculate properties of structure with potential')
    parser.set_defaults(func=handle_subcommand_test_properties)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-s', '--structure', help='base structure', type=is_file_type, required=True)
    parser.add_argument('--property', action='append', choices=['lattice', 'elastic', 'static'], help='choose properties to test default is all')
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('--num-workers', default=1, type=int, help='number md calculators to use')


def add_subcommand_test_relax(subparsers):
    parser = subparsers.add_parser('relax', help='calculate relaxed structure from potential')
    parser.set_defaults(func=handle_subcommand_test_relax)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-s', '--structure', help='base structure', type=is_file_type, required=True)
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write relaxed structure', required=True)


def add_subcommand_test_training(subparsers):
    parser = subparsers.add_parser('training', help='plot difference between md and training set preditions')
    parser.set_defaults(func=handle_subcommand_test_training)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('-t', '--training', help='training set to use for comparison', type=is_file_type, required=True)
    parser.add_argument('plot', help='what to plot', choices=['energy', 'forces', 'stress'])
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')
    parser.add_argument('--cache', default='~/.cache/dftfit/cache.db', help='dft cache', type=is_file_type)
    parser.add_argument('--hide', dest='show', action='store_false', help='do not show plot')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write visualization to')


def add_subcommand_test_radial(subparsers):
    parser = subparsers.add_parser('radial', help='plot the radial pair distributions for each atom pair type')
    parser.set_defaults(func=handle_subcommand_test_radial)
    parser.add_argument('-t', '--training', help='training set to use for comparison', type=is_file_type, required=True)
    parser.add_argument('--distance', default=10.0, help='distance to calculate radial distribution function', type=float)
    parser.add_argument('--cache', default='~/.cache/dftfit/cache.db', help='dft cache', type=is_file_type)
    parser.add_argument('--hide', dest='show', action='store_false', help='do not show plot')
    parser.add_argument('-o', '--output-filename', type=is_not_file_type, help='filename to write visualization to')


def add_subcommand_test_pair(subparsers):
    parser = subparsers.add_parser('pair', help='plot potential energy for each atom pair type in potential')
    parser.set_defaults(func=handle_subcommand_test_pair)
    parser.add_argument('-p', '--potential', help='potential filename in in yaml/json format', type=is_file_type, required=True)
    parser.add_argument('--min', default=0.1, type=float, help='minimum distance')
    parser.add_argument('--max', default=10.0, type=float, help='maximum distance to test potential')
    parser.add_argument('--samples', default=100, type=int, help='number of samples of potential')
    parser.add_argument('--software', default='lammps', help='md calculator to use')
    parser.add_argument('--command', help='md calculator command has sensible defaults')



def get_structure(filename):
    if filename.lower().endswith('.cif'):
        structure = CifParser(filename).get_structures()[0]
        lattice = pmg.Lattice.from_parameters(*structure.lattice.abc, *structure.lattice.angles)
        return pmg.Structure(lattice, structure.species, structure.frac_coords, coords_are_cartesian=False)
    elif filename.lower().endswith('poscar'):
        return Poscar.from_file(filename).structure
    else:
        raise ValueError('Cannot determine file type from filename [.cif, poscar]')


def handle_subcommand_test_properties(args):
    properties = set(args.property) if args.property else {'elastic', 'lattice', 'static'}
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=args.num_workers)
    potential = Potential.from_file(args.potential)
    structure = get_structure(args.structure)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    if 'lattice' in properties:
        old_lattice, new_lattice = predict.lattice_constant(structure, potential)
        print('\nLattice Constants:')
        print('        a: {:6.3f}    b: {:6.3f}     c: {:6.3f}'.format(*new_lattice.abc))
        print('    alpha: {:6.3f} beta: {:6.3f} gamma: {:6.3f}'.format(*new_lattice.angles))
    if 'elastic' in properties:
        elastic = predict.elastic_constant(structure, potential)
        print('\nElastic:')
        print_elastic_information(elastic)

    if 'static' in properties:
        static = predict.static(structure, potential)
        print('\nStatic:')
        print('    Energy: [eV]')
        print('      {:16.3f}'.format(static['energy']))
        print('    Forces: [eV/Angstrom]')
        for row in static['forces']:
            print('      {:16.3f} {:16.3f} {:16.3f}'.format(*row))
        print('    Stress: [bars]')
        for row in static['stress']:
            print('      {:16.3f} {:16.3f} {:16.3f}'.format(*row))


def handle_subcommand_test_relax(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)
    structure = get_structure(args.structure)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    sga = SpacegroupAnalyzer(structure)
    conventional_structure = sga.get_conventional_standard_structure()
    old_lattice, new_lattice = predict.lattice_constant(conventional_structure, potential)
    equilibrium_structure = Structure(
        new_lattice,
        [s.specie.element for s in conventional_structure.sites],
        [s.frac_coords for s in conventional_structure.sites])
    equilibrium_structure.to(filename=args.output_filename)


def handle_subcommand_test_training(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    cache_filename = os.path.expanduser(args.cache)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)
    training = Training.from_file(args.training, cache_filename=cache_filename)

    import warnings
    warnings.filterwarnings("ignore") # yes I have sinned

    md_calculations = []
    for dft_calculation in training.calculations:
        result = predict.static(dft_calculation.structure, potential)
        md_calculations.append(MinimalMDReader(
            forces=result['forces'],
            stress=result['stress'],
            energy=result['energy'],
            structure=dft_calculation.structure))
    visualize_single_calculation(training.calculations, md_calculations, plot=args.plot, show=args.show, filename=args.output_filename)


def handle_subcommand_test_radial(args):
    cache_filename = os.path.expanduser(args.cache)
    training = Training.from_file(args.training, cache_filename=cache_filename)
    visualize_radial_pair_distribution(training.calculations, distance=args.distance, show=args.show, filename=args.output_filename)


def handle_subcommand_test_pair(args):
    default_commands = {
        'lammps': 'lammps'
    }
    command = args.command if args.command else default_commands.get(args.software)
    predict = Predict(calculator=args.software, command=command, num_workers=1)
    potential = Potential.from_file(args.potential)

    seperations = np.linspace(args.min, args.max, args.samples)
    pair_energies = {}
    for element_a, element_b in itertools.combinations_with_replacement(potential.elements, 2):
        energies = predict.pair(element_a, element_b, potential, seperations)
        pair_energies['%s-%s' % (element_a, element_b)] = energies
    visualize_pair_energies(seperations, pair_energies)



import functools
import logging

import pygmo

from .problem import DFTFITSingleProblem, DFTFITMultiProblem

logger = logging.getLogger(__name__)


available_algorithms = {
    'pygmo.de': (pygmo.de, 'S'), # S-U
    'pygmo.sade': (pygmo.sade, 'S'), # S-U
    'pygmo.de1220': (pygmo.de1220, 'S'), # S-U
    'pygmo.pso': (pygmo.pso, 'S'), # S-U
    'pygmo.sea': (pygmo.sea, 'S'), # S-U
    'pygmo.sga': (pygmo.sga, 'S'), # S-U
    # 'pygmo.simulated_annealing': simulated_annealing # api does not match others
    'pygmo.bee_colony': (pygmo.bee_colony, 'S'), # S-U
    'pygmo.cmaes': (pygmo.cmaes, 'S'), # S-U population > 5 and % 4
    'pygmo.nsga2': (pygmo.nsga2, 'M'), # M-U
    'pygmo.moead': (pygmo.moead, 'M'), # M-U
    'nlopt.cobyla': (functools.partial(pygmo.nlopt, solver='cobyla'), 'S'), # S-U
    'nlopt.bobyqa': (functools.partial(pygmo.nlopt, solver='bobyqa'), 'S'), # S-U
    'nlopt.newuoa': (functools.partial(pygmo.nlopt, solver='newuoa'), 'S'), # S-U
    'nlopt.sbplx': (functools.partial(pygmo.nlopt, solver='sbplx'), 'S'), # S-U
}


class Optimize:
    def __init__(self, potential, training,
                 dbm=None, db_write_interval=10,                      # database
                 algorithm='pygmo.de', algorithm_kwargs=None,         # algorithm
                 features=None, weights=None, problem_kwargs=None,    # problem
                 run_id=None):

        self.algorithm_name = algorithm
        if self.algorithm_name not in available_algorithms:
            raise ValueError(f'algorithm {self.algorithm_name} not available')

        _problem_kwargs = {
            'potential': potential,
            'training': training,
            'dbm': dbm, 'db_write_interval': db_write_interval,
            'features': features, 'weights': weights,
            'run_id': run_id,
            **problem_kwargs
        }
        if available_algorithms[self.algorithm_name][1] == 'S':
            internal_problem = DFTFITSingleProblem(**_problem_kwargs)
        else:
            internal_problem = DFTFITMultiProblem(**_problem_kwargs)
        self._internal_problem = internal_problem
        self._problem = pygmo.problem(internal_problem)
        self.algorithm_kwargs = algorithm_kwargs or {}

    def population(self, size, seed=None):
        return pygmo.population(self._problem, size, seed=seed)

    def optimize(self, population, steps, seed=None):
        algorithm_constructor = available_algorithms[self.algorithm_name][0]
        if 'nlopt' in self.algorithm_name: # nlopt algorithms called differently.
            _algorithm = algorithm_constructor()
            _algorithm.maxeval = steps
            self._algorithm = pygmo.algorithm(_algorithm)
        else:
            self._algorithm = pygmo.algorithm(algorithm_constructor(gen=steps, seed=seed, **self.algorithm_kwargs))
        logger.info('(algorithm) using %s algorithm with steps: %d seed: %d' % (self.algorithm_name, steps, seed))

        results = self._algorithm.evolve(population)
        self._internal_problem.finalize()
        return results



def print_elastic_information(elastic):
    print('Stiffness Tensor')
    for row in elastic.voigt:
        print('{:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f} {:+8.1f}\n'.format(*row))
    print('\n')
    print('Shear Modulus G_V', elastic.g_voigt)
    print('Shear Modulus G_R', elastic.g_reuss)
    print('Shear Modulus G_vrh', elastic.g_vrh)

    print('Bulk Modulus K_V', elastic.k_voigt)
    print('Bulk Modulus K_R', elastic.k_reuss)
    print('Bulk Modulus K_vrh', elastic.k_vrh)

    print('Elastic Anisotropy', elastic.universal_anisotropy)
    print('Poisons Ration', elastic.homogeneous_poisson)


def apply_structure_operations(structure, operations, tollerance=0.1):
    for operation in operations:
        cart_coords = operation['position']
        element = operation['element']
        if operation['opp'] == '+':
            structure.append(element, cart_coords)
        elif operation['opp'] == '-':
            sites = structure.get_sites_in_sphere(cart_coords, tollerance)
            if len(sites) != 1:
                raise ValueError('found %d sites at %s with radius %f needed only one' % (len(sites), cart_coords, tollerance))
            site = sites[0][0]
            structure.remove_sites([structure.index(site)])



import asyncio
import os
import json
import math
import copy
import logging

import numpy as np
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Lattice, Structure, Element
from pymatgen.analysis.elasticity import DeformedStructureSet, ElasticTensor, Stress, Strain
from pymatgen.util.coord import pbc_diff

from ..io.lammps import LammpsLocalMDCalculator
from ..io.lammps_cython import LammpsCythonMDCalculator
from .utils import apply_structure_operations

logger = logging.getLogger(__name__)


def load_lammps_set(config_filename):
    from pmg_lammps.sets import MODULE_DIR
    from pmg_lammps.inputs import LammpsScript

    with open(os.path.join(MODULE_DIR, 'sets', config_filename + ".json")) as f:
        return json.load(f, object_pairs_hook=LammpsScript)


class Predict:
    def __init__(self, calculator='lammps_cython', loop=None, **kwargs):
        calculator_mapper = {
            'lammps': LammpsLocalMDCalculator,
            'lammps_cython': LammpsCythonMDCalculator
        }
        self.calculator_type = calculator
        self.calculator = calculator_mapper[calculator](**kwargs)
        self.loop = loop or asyncio.get_event_loop()
        self.loop.run_until_complete(self.calculator.create())

    def conventional_structure(self, structure):
        sga = SpacegroupAnalyzer(structure)
        return sga.get_conventional_standard_structure()

    def material_properties(self, structure, potential, properties=None):
        properties = properties or {'lattice_constants'}
        raise NotImplementedError()

    def static(self, structure, potential):
        if self.calculator_type == 'lammps':
            kwargs = {'lammps_set': load_lammps_set('static')}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': ['run 0']}

        async def calculate():
            future = await self.calculator.submit(
                structure, potential,
                properties={'forces', 'stress', 'energy'},
                **kwargs)
            await future
            return future.result()
        result = self.loop.run_until_complete(calculate())
        return {
            'energy': result['results']['energy'],
            'stress': np.array(result['results']['stress']) * 1e-4,  # convert to GPa
            'forces': np.array(result['results']['forces'])
        }

    def pair(self, element_a, element_b, potential, separations):
        max_r = np.max(separations)
        lattice = Lattice.from_parameters(10*max_r, 10*max_r, 10*max_r, 90, 90, 90)

        if self.calculator_type == 'lammps':
            kwargs = {'lammps_set': load_lammps_set('static')}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': ['run 0']}

        async def calculate():
            futures = []
            for sep in separations:
                coord_a = (lattice.a*0.5-(sep/2), lattice.b*0.5, lattice.c*0.5)
                coord_b = (lattice.a*0.5+(sep/2), lattice.b*0.5, lattice.c*0.5)
                structure = Structure(
                    lattice,
                    [element_a, element_b],
                    [coord_a, coord_b], coords_are_cartesian=True)
                futures.append(await self.calculator.submit(
                    structure, potential,
                    properties={'energy'},
                    **kwargs))
            return await asyncio.gather(*futures)
        results = self.loop.run_until_complete(calculate())
        return np.array([r['results']['energy'] for r in results])

    def three_body(self, element_a, element_b, potential, separation, angles):
        max_r = separation
        lattice = Lattice.from_parameters(10*max_r, 10*max_r, 10*max_r, 90, 90, 90)

        if self.calculator_type == 'lammps':
            kwargs = {'lammps_set': load_lammps_set('static')}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': ['run 0']}

        async def calculate():
            futures = []
            for angle in angles:
                coord_a = (lattice.a*0.5 + separation, lattice.b*0.5, lattice.c*0.5)
                coord_b = (lattice.a*0.5, lattice.b*0.5, lattice.c*0.5) # b in center
                coord_c = (lattice.a*0.5 + (math.cos(angle) * separation), lattice.b*0.5 + (math.sin(angle) * separation), lattice.c*0.5)
                structure = Structure(
                    lattice,
                    [element_b, element_a, element_b],
                    [coord_a, coord_b, coord_c], coords_are_cartesian=True)
                futures.append(await self.calculator.submit(
                    structure, potential,
                    properties={'energy'},
                    **kwargs))
            return await asyncio.gather(*futures)
        results = self.loop.run_until_complete(calculate())
        return np.array([r['results']['energy'] for r in results])

    def lattice_constant(self, structure, potential, supercell=(1, 1, 1), etol=1e-6, ftol=1e-6, nsearch=2000, neval=10000):
        conventional_structure = self.conventional_structure(structure)

        if self.calculator_type == 'lammps':
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['minimize'] = '%f %f %d %d' % (etol, ftol, nsearch, neval)
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'fix 1 all box/relax iso 0.0 vmax 0.001',
                'min_style cg',
                'minimize %f %f %d %d' % (etol, ftol, nsearch, neval)
            ]}

        async def calculate():
            future = await self.calculator.submit(
                conventional_structure * supercell, potential,
                properties={'lattice'},
                **kwargs)
            await future
            return future.result()

        result = self.loop.run_until_complete(calculate())
        lattice = Lattice(result['results']['lattice'])
        return conventional_structure.lattice, Lattice(lattice.matrix / np.array(supercell))

    def elastic_constant(self, structure, potential, supercell=(1, 1, 1),
                         nd=0.01, ns=0.05, num_norm=4, num_shear=4,
                         etol=1e-6, ftol=1e-6, nsearch=2000, neval=10000):
        norm_strains = np.linspace(-nd, nd, num_norm).tolist()
        shear_strains = np.linspace(-ns, ns, num_shear).tolist()
        conventional_structure = self.conventional_structure(structure)
        deformation_set = DeformedStructureSet(conventional_structure * supercell,
                                               norm_strains=norm_strains,
                                               shear_strains=shear_strains)
        strains = []
        stresses = []

        if self.calculator_type == 'lammps':
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f %d %d' % (etol, ftol, nsearch, neval)
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'min_style cg',
                'minimize %f %f %d %d' % (etol, ftol, nsearch, neval)
            ]}

        async def calculate():
            futures = []
            for deformation, deformed_structure in zip(deformation_set.deformations, deformation_set):
                strains.append(Strain.from_deformation(deformation))
                futures.append(await self.calculator.submit(
                    deformed_structure, potential,
                    properties={'stress'},
                    **kwargs))
            for result in await asyncio.gather(*futures):
                stress = Stress(np.array(result['results']['stress']) * 1e-4) # Convert to GPa
                stresses.append(stress)

        self.loop.run_until_complete(calculate())
        return ElasticTensor.from_independent_strains(strains, stresses, Stress(np.zeros((3, 3))))

    def point_defects(self, structure, potential, point_defect_schemas, supercell=(1, 1, 1), etol=1e-6, ftol=1e-6, nsearch=2000, neval=10000):
        """ Calculate the energy of each defect.

        structure is assumed to be the relaxed structure. This can take a long time (1-2 minutes per relaxation)
        """
        if self.calculator_type == 'lammps':
            relax_lammps_script = load_lammps_set('relax')
            relax_lammps_script['fix'] = []
            relax_lammps_script['minimize'] = '%f %f %d %d' % (etol, ftol, nsearch, neval)
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'min_style cg',
                'minimize %f %f %d %d' % (etol, ftol, nsearch, neval)
            ]}

        energies = {}
        point_defect_schemas = copy.deepcopy(point_defect_schemas)
        for point_defect_name, operations in point_defect_schemas.items():
            base_structure = structure.copy()
            for operation in operations:
                cart_coords = base_structure.lattice.get_cartesian_coords(operation['position'])
                operation['position'] = cart_coords

            base_structure = base_structure * supercell
            apply_structure_operations(base_structure, operations)

            async def calculate():
                future = await self.calculator.submit(
                    base_structure, potential,
                    properties={'energy', 'timesteps'},
                    **kwargs)
                await future
                return future.result()

            logger.info('starting calculation (point defect): %s' % point_defect_name)
            result = self.loop.run_until_complete(calculate())
            num_eval = result['results']['timesteps']
            logger.info('finished calculation (point defect): %s in evals: %d' % (point_defect_name, num_eval))
            energies[point_defect_name] = result['results']['energy']
        return energies

    def displacement_energies(self, structure, potential, displacement_energies_schema, supercell=(1, 1, 1), tollerance=0.1, max_displacement_energy=75, resolution=1, num_steps=1000, site_radius=0.5, timestep=0.001):
        """ Calculate displacement energy for each atom.

        Uses bisection method to determine displacement energy.
        """
        def ev2Aps(Z, energy):
            # sqrt((2 * energy[eV] [J/eV]) / (amu [g/mole] [kg/g])) * [m/s] [A/ps]
            return math.sqrt((2 * energy * 1.6021766208e-19) / (Z / (6.02214085e23 * 1e3))) * 1e-2

        if self.calculator_type == 'lammps':
            logger.warning('"lammps" calculator is depriciated use "lammps_cython" cannot promise working')
            relax_lammps_script = load_lammps_set('nve')
            relax_lammps_script['thermo'] = []
            relax_lammps_script
            relax_lammps_script['timestep'] = timestep # fs
            relax_lammps_script['run'] = num_steps
            kwargs = {'lammps_set': relax_lammps_script}
        elif self.calculator_type == 'lammps_cython':
            kwargs = {'lammps_additional_commands': [
                'timestep %f' % timestep,
                'velocity all zero linear',
                'fix 1 all nve',
                'run %d' % num_steps
            ]}

        energies = {}
        displacement_energies_schema = displacement_energies_schema.copy()
        for displacement_energy_name, d in displacement_energies_schema.items():
            base_structure = structure.copy()
            v = base_structure.lattice.get_cartesian_coords(d['direction'])
            cart_coords = base_structure.lattice.get_cartesian_coords(d['position'])
            base_structure = base_structure * supercell
            site = base_structure.get_sites_in_sphere(cart_coords, tollerance)[0][0]
            original_positions = base_structure.cart_coords
            original_frac_positions = base_structure.lattice.get_fractional_coords(original_positions)
            index = base_structure.index(site)

            min_energy, max_energy = 0.0, max_displacement_energy
            guess_energy = None
            while abs(max_energy - min_energy) > resolution:
                guess_energy = (max_energy - min_energy) / 2 + min_energy
                velocity = (v / np.linalg.norm(v)) * ev2Aps(Element(d['element']).atomic_mass, guess_energy)
                velocities = np.zeros((len(base_structure), 3))
                velocities[index] = velocity
                base_structure.add_site_property('velocities', velocities)

                async def calculate():
                    future = await self.calculator.submit(
                        base_structure, potential,
                        properties={'positions', 'initial_positions'},
                        **kwargs)
                    await future
                    return future.result()

                print('starting calculation (displacement energy): %s ion %s velocity: %f [eV] %f [A/ps]' % (displacement_energy_name, d['element'], guess_energy, ev2Aps(Element(d['element']).atomic_mass, guess_energy)))
                result = self.loop.run_until_complete(calculate())
                initial_frac_positions = base_structure.lattice.get_fractional_coords(result['results']['initial_positions'])
                final_frac_positions = base_structure.lattice.get_fractional_coords(result['results']['positions'])
                displacements = np.linalg.norm(
                    base_structure.lattice.get_cartesian_coords(
                        pbc_diff(final_frac_positions, initial_frac_positions)), axis=1)
                is_original_state = np.all(displacements < site_radius)
                print('finished calculation (displacement energy): %s resulted in ground_state (%s) max displacment %f [A] median %f [A] min %f [A]' % (displacement_energy_name, is_original_state, np.max(displacements), np.median(displacements), np.min(displacements)))
                if is_original_state:
                    min_energy = guess_energy
                else:
                    max_energy = guess_energy

            energies[displacement_energy_name] = guess_energy
        return energies
